@@ -12,10 +12,6 @@
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  */
 #pragma once
 
@@ -36,6 +32,8 @@ struct bitvec;
 
 #ifdef __cplusplus
 
+#include "rlc_window_dl.h"
+
 class Decoding {
 public:
 	/* represents (parts) LLC PDUs within one RLC Data block */
