@@ -31,6 +31,7 @@ extern "C" {
 #include <osmocom/core/timer.h>
 #include <osmocom/core/linuxlist.h>
 #include <osmocom/core/rate_ctr.h>
+#include <osmocom/core/use_count.h>
 
 #include <osmocom/gsm/protocol/gsm_23_003.h>
 #include <osmocom/gsm/gsm48.h>
@@ -50,14 +51,8 @@ struct gprs_rlcmac_bts;
 struct gprs_rlcmac_trx;
 struct GprsMs;
 
-struct gpr_ms_callback {
-	void (*ms_idle)(struct GprsMs *);
-	void (*ms_active)(struct GprsMs *);
-};
-
 struct GprsMs {
 	struct llist_head list; /* list of all GprsMs */
-	struct gpr_ms_callback cb;
 	bool app_info_pending;
 
 	struct gprs_rlcmac_bts *bts;
@@ -83,10 +78,10 @@ struct GprsMs {
 	struct gprs_llc_queue llc_queue;
 	struct osmo_timer_list llc_timer;
 
-	bool is_idle;
-	int ref;
-	struct osmo_timer_list timer;
-	unsigned delay;
+	struct osmo_use_count use_count;
+	struct osmo_timer_list release_timer;
+	/* Time at which MS became idle and waiting to be released by release_timer: */
+	struct timeval tv_idle_start;
 
 	int64_t last_cs_not_low;
 
@@ -101,14 +96,12 @@ struct GprsMs {
 	struct nacc_fsm_ctx *nacc;
 };
 
-struct GprsMs *ms_alloc(struct gprs_rlcmac_bts *bts, uint32_t tlli);
+struct GprsMs *ms_alloc(struct gprs_rlcmac_bts *bts, const char *use_ref);
 
 struct gprs_rlcmac_pdch *ms_first_common_ts(const struct GprsMs *ms);
 void ms_set_first_common_ts(struct GprsMs *ms, struct gprs_rlcmac_pdch *pdch);
 void ms_set_reserved_slots(struct GprsMs *ms, struct gprs_rlcmac_trx *trx,
 			   uint8_t ul_slots, uint8_t dl_slots);
-struct GprsMs *ms_ref(struct GprsMs *ms);
-void ms_unref(struct GprsMs *ms);
 void ms_set_mode(struct GprsMs *ms, enum mcs_kind mode);
 void ms_set_ms_class(struct GprsMs *ms, uint8_t ms_class_);
 void ms_set_egprs_ms_class(struct GprsMs *ms, uint8_t ms_class_);
@@ -131,7 +124,6 @@ void ms_detach_tbf(struct GprsMs *ms, struct gprs_rlcmac_tbf *tbf);
 void ms_set_tlli(struct GprsMs *ms, uint32_t tlli);
 bool ms_confirm_tlli(struct GprsMs *ms, uint32_t tlli);
 void ms_set_imsi(struct GprsMs *ms, const char *imsi);
-uint16_t ms_paging_group(struct GprsMs *ms);
 
 void ms_update_l1_meas(struct GprsMs *ms, const struct pcu_l1_meas *meas);
 
@@ -142,8 +134,6 @@ static inline struct gprs_rlcmac_dl_tbf *ms_dl_tbf(const struct GprsMs *ms) {ret
 const char *ms_name(const struct GprsMs *ms);
 char *ms_name_buf(const struct GprsMs *ms, char *buf, unsigned int buf_size);
 
-void ms_set_callback(struct GprsMs *ms, struct gpr_ms_callback *cb);
-
 int ms_nacc_start(struct GprsMs *ms, Packet_Cell_Change_Notification_t *notif);
 bool ms_nacc_rts(const struct GprsMs *ms);
 struct msgb *ms_nacc_create_rlcmac_msg(struct GprsMs *ms, struct gprs_rlcmac_tbf *tbf,
@@ -158,7 +148,8 @@ int ms_append_llc_dl_data(struct GprsMs *ms, uint16_t pdu_delay_csec, const uint
 
 static inline bool ms_is_idle(const struct GprsMs *ms)
 {
-	return !ms->ul_tbf && !ms->dl_tbf && !ms->ref && llist_empty(&ms->old_tbfs);
+	return !ms->ul_tbf && !ms->dl_tbf &&
+		llist_empty(&ms->old_tbfs);
 }
 
 static inline struct gprs_llc_queue *ms_llc_queue(struct GprsMs *ms)
@@ -166,11 +157,30 @@ static inline struct gprs_llc_queue *ms_llc_queue(struct GprsMs *ms)
 	return &ms->llc_queue;
 }
 
+/* Function used in code where a ul_tbf related event occurs and hence its state
+ * changes, which in turn may change the entire MS state (eg becoming reachable
+ * over PCH or PACCH) and the caller may want to know if it is a good time to
+ * assign a DL TBF (and whether we have DL data to send) */
 static inline bool ms_need_dl_tbf(struct GprsMs *ms)
 {
-	if (ms_dl_tbf(ms) != NULL &&
-	    tbf_state((const struct gprs_rlcmac_tbf *)ms_dl_tbf(ms)) != TBF_ST_WAIT_RELEASE)
-		return false;
+	struct gprs_rlcmac_dl_tbf *dl_tbf = ms_dl_tbf(ms);
+	if (dl_tbf) {
+		switch (tbf_state(dl_tbf_as_tbf(dl_tbf))) {
+		case TBF_ST_NEW:
+		case TBF_ST_ASSIGN:
+		case TBF_ST_FLOW:
+		case TBF_ST_FINISHED:
+		case TBF_ST_WAIT_RELEASE:
+			return false; /* TBF in use, hence no need for new DL TBF */
+		case TBF_ST_WAIT_REUSE_TFI:
+		case TBF_ST_RELEASING:
+			/* TBF cannot be used to send further data, a new one
+			 * may be needed, check below */
+			break;
+		default:
+			OSMO_ASSERT(0);
+		}
+	}
 
 	return llc_queue_size(ms_llc_queue(ms)) > 0;
 }
@@ -226,11 +236,6 @@ static inline enum mcs_kind ms_mode(const struct GprsMs *ms)
 	return ms->mode;
 }
 
-static inline void ms_set_timeout(struct GprsMs *ms, unsigned secs)
-{
-	ms->delay = secs;
-}
-
 static inline unsigned ms_nack_rate_dl(const struct GprsMs *ms)
 {
 	return ms->nack_rate_dl;
@@ -251,6 +256,12 @@ static inline struct gprs_rlcmac_trx *ms_current_trx(const struct GprsMs *ms)
 	return ms->current_trx;
 }
 
+#define MS_USE_TBF "tbf"
+#define ms_ref(ms, use) \
+	OSMO_ASSERT(osmo_use_count_get_put(&(ms)->use_count, use, 1) == 0)
+#define ms_unref(ms, use) \
+	OSMO_ASSERT(osmo_use_count_get_put(&(ms)->use_count, use, -1) == 0)
+
 #define LOGPMS(ms, category, level, fmt, args...) \
 	LOGP(category, level, "%s " fmt, ms_name(ms), ## args)
 
