@@ -26,7 +26,6 @@
 #include "tbf_ul.h"
 #include "gprs_ms.h"
 #include "gprs_debug.h"
-#include "gprs_ms_storage.h"
 #include "pcu_utils.h"
 #include "gprs_bssgp_pcu.h"
 #include "pcu_l1_if.h"
@@ -43,7 +42,7 @@ extern "C" {
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/utils.h>
 #include <osmocom/vty/vty.h>
-#include <osmocom/gprs/protocol/gsm_04_60.h>
+#include <osmocom/gsm/protocol/gsm_44_060.h>
 #include <osmocom/gsm/l1sap.h>
 #include <osmocom/core/fsm.h>
 }
@@ -81,7 +80,8 @@ static int bts_handle_rach(struct gprs_rlcmac_bts *bts, uint16_t ra, uint32_t Fn
 		.ra = ra,
 		.trx_nr = 0,
 		.ts_nr = 0,
-		.rfn = Fn,
+		.rfn = fn2rfn(Fn),
+		.fn = Fn,
 		.qta = qta,
 	};
 
@@ -123,23 +123,22 @@ static void test_tbf_tlli_update()
 	/*
 	 * Make a uplink and downlink allocation
 	 */
-	ms = bts_alloc_ms(bts, 0, 0);
-	gprs_rlcmac_tbf *dl_tbf = dl_tbf_alloc(bts,
-						ms, 0, false);
+	ms = ms_alloc(bts, NULL);
+	OSMO_ASSERT(ms_new_dl_tbf_assigned_on_pch(ms) == 0);
+	gprs_rlcmac_tbf *dl_tbf = ms_dl_tbf(ms);
 	OSMO_ASSERT(dl_tbf != NULL);
+	OSMO_ASSERT(dl_tbf->ms() == ms);
 	ms_confirm_tlli(ms, 0x2342);
 	dl_tbf->set_ta(4);
-	OSMO_ASSERT(ms_dl_tbf(ms) == dl_tbf);
-	OSMO_ASSERT(dl_tbf->ms() == ms);
 
-	gprs_rlcmac_tbf *ul_tbf = ul_tbf_alloc(bts,
-						   ms, 0, false);
+	OSMO_ASSERT(ms_new_ul_tbf_assigned_pacch(ms, 0));
+
+	gprs_rlcmac_tbf *ul_tbf = ms_ul_tbf(ms);
 	OSMO_ASSERT(ul_tbf != NULL);
-	ms_update_announced_tlli(ms, 0x2342);
-	OSMO_ASSERT(ms_ul_tbf(ms) == ul_tbf);
 	OSMO_ASSERT(ul_tbf->ms() == ms);
+	ms_update_announced_tlli(ms, 0x2342);
 
-	OSMO_ASSERT(bts_ms_by_tlli(bts, 0x2342, GSM_RESERVED_TMSI) == ms);
+	OSMO_ASSERT(bts_get_ms_by_tlli(bts, 0x2342, GSM_RESERVED_TMSI) == ms);
 
 	/*
 	 * Now check.. that DL changes and that the timing advance
@@ -148,20 +147,20 @@ static void test_tbf_tlli_update()
 	ms_confirm_tlli(dl_tbf->ms(), 0x4232);
 
 	/* It is still there, since the new TLLI has not been used for UL yet */
-	ms_new = bts_ms_by_tlli(bts, 0x2342, GSM_RESERVED_TMSI);
+	ms_new = bts_get_ms_by_tlli(bts, 0x2342, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms == ms_new);
 
-	ms_new = bts_ms_by_tlli(bts, 0x4232, GSM_RESERVED_TMSI);
+	ms_new = bts_get_ms_by_tlli(bts, 0x4232, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms == ms_new);
 	OSMO_ASSERT(ms_dl_tbf(ms) == dl_tbf);
 	OSMO_ASSERT(ms_ul_tbf(ms) == ul_tbf);
 
 	/* Now use the new TLLI for UL */
 	ms_update_announced_tlli(ms, 0x4232);
-	ms_new = bts_ms_by_tlli(bts, 0x2342, GSM_RESERVED_TMSI);
+	ms_new = bts_get_ms_by_tlli(bts, 0x2342, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms_new == NULL);
 
-	ms_new = bts_ms_by_tlli(bts, 0x4232, GSM_RESERVED_TMSI);
+	ms_new = bts_get_ms_by_tlli(bts, 0x4232, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms_new != NULL);
 	OSMO_ASSERT(ms_ta(ms_new) == 4);
 
@@ -198,34 +197,6 @@ static void setup_bts(struct gprs_rlcmac_bts *bts, uint8_t ts_no, uint8_t cs = 1
 	bts_set_current_frame_number(bts, DUMMY_FN);
 }
 
-static gprs_rlcmac_dl_tbf *create_dl_tbf(struct gprs_rlcmac_bts *bts, uint8_t ms_class,
-	uint8_t egprs_ms_class, uint8_t *trx_no_)
-{
-	int tfi;
-	uint8_t trx_no;
-	GprsMs *ms;
-	gprs_rlcmac_dl_tbf *dl_tbf;
-
-	ms = bts_alloc_ms(bts, ms_class, egprs_ms_class);
-
-	tfi = bts_tfi_find_free(bts, GPRS_RLCMAC_DL_TBF, &trx_no, -1);
-	OSMO_ASSERT(tfi >= 0);
-	dl_tbf = dl_tbf_alloc(bts, ms, trx_no, true);
-	OSMO_ASSERT(dl_tbf);
-	dl_tbf->set_ta(0);
-	check_tbf(dl_tbf);
-
-	/* "Establish" the DL TBF */
-	osmo_fsm_inst_dispatch(dl_tbf->dl_ass_fsm.fi, TBF_DL_ASS_EV_SCHED_ASS, NULL);
-	osmo_fsm_inst_dispatch(dl_tbf->state_fi, TBF_EV_ASSIGN_ADD_CCCH, NULL);
-	osmo_fsm_inst_dispatch(dl_tbf->state_fi, TBF_EV_ASSIGN_ACK_PACCH, NULL);
-	check_tbf(dl_tbf);
-
-	*trx_no_ = trx_no;
-
-	return dl_tbf;
-}
-
 static unsigned fn_add_blocks(unsigned fn, unsigned blocks)
 {
 	unsigned bn = fn2bn(fn) + blocks;
@@ -252,6 +223,36 @@ static void request_dl_rlc_block(struct gprs_rlcmac_tbf *tbf,
 	request_dl_rlc_block(tbf->bts, tbf->control_ts, fn, block_nr);
 }
 
+static gprs_rlcmac_dl_tbf *create_dl_tbf(struct gprs_rlcmac_bts *bts, uint8_t ms_class,
+	uint8_t egprs_ms_class, uint8_t *trx_no_)
+{
+	int tfi;
+	uint8_t trx_no;
+	GprsMs *ms;
+	gprs_rlcmac_dl_tbf *dl_tbf;
+
+	ms = ms_alloc(bts, NULL);
+	ms_set_ms_class(ms, ms_class);
+	ms_set_egprs_ms_class(ms, egprs_ms_class);
+
+	tfi = bts_tfi_find_free(bts, GPRS_RLCMAC_DL_TBF, &trx_no, -1);
+	OSMO_ASSERT(tfi >= 0);
+	OSMO_ASSERT(ms_new_dl_tbf_assigned_on_pch(ms) == 0);
+	dl_tbf = ms_dl_tbf(ms);
+	OSMO_ASSERT(dl_tbf);
+	dl_tbf->set_ta(0);
+	check_tbf(dl_tbf);
+
+	/* "Establish" the DL TBF */
+	osmo_fsm_inst_dispatch(dl_tbf->state_fi, TBF_EV_ASSIGN_ADD_CCCH, NULL);
+	osmo_fsm_inst_dispatch(dl_tbf->state_fi, TBF_EV_ASSIGN_READY_CCCH, NULL);
+	check_tbf(dl_tbf);
+
+	*trx_no_ = trx_no;
+
+	return dl_tbf;
+}
+
 enum test_tbf_final_ack_mode {
 	TEST_MODE_STANDARD,
 	TEST_MODE_REVERSE_FREE
@@ -326,6 +327,13 @@ static void test_tbf_final_ack(enum test_tbf_final_ack_mode test_mode)
 	OSMO_ASSERT(dl_tbf->have_data());
 	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
 
+	do {
+		/* Request to send one block */
+		request_dl_rlc_block(dl_tbf, &fn, &block_nr);
+	} while (dl_tbf->state_is(TBF_ST_FLOW));
+
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FINISHED));
+
 	/* Queue a final ACK */
 	memset(rbb, 0, sizeof(rbb));
 	/* Receive a final ACK */
@@ -333,26 +341,29 @@ static void test_tbf_final_ack(enum test_tbf_final_ack_mode test_mode)
 
 	/* Clean up and ensure tbfs are in the correct state */
 	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_WAIT_RELEASE));
+
+	/* Now append new data, a new DL TBF should be created and assigned through the old one due to T3192: */
+	ms_append_llc_dl_data(dl_tbf->ms(), 1000, llc_data, sizeof(llc_data));
 	new_tbf = ms_dl_tbf(ms);
 	check_tbf(new_tbf);
 	OSMO_ASSERT(new_tbf != dl_tbf);
 	OSMO_ASSERT(new_tbf->tfi() == 1);
 	check_tbf(dl_tbf);
 	if (test_mode == TEST_MODE_REVERSE_FREE) {
-		ms_ref(ms);
+		ms_ref(ms, __func__);
 		tbf_free(new_tbf);
 		OSMO_ASSERT(ms_dl_tbf(ms) == NULL);
 		check_tbf(dl_tbf);
 		tbf_free(dl_tbf);
-		ms_unref(ms);
+		ms_unref(ms, __func__);
 	} else {
-		ms_ref(ms);
+		ms_ref(ms, __func__);
 		tbf_free(dl_tbf);
 		OSMO_ASSERT(ms_dl_tbf(ms) == new_tbf);
 		check_tbf(new_tbf);
 		tbf_free(new_tbf);
 		OSMO_ASSERT(ms_dl_tbf(ms) == NULL);
-		ms_unref(ms);
+		ms_unref(ms, __func__);
 	}
 
 	TALLOC_FREE(the_pcu);
@@ -450,39 +461,39 @@ static void test_tbf_imsi()
 	ms_confirm_tlli(dl_tbf[1]->ms(), 0xf1000002);
 
 	ms_set_imsi(dl_tbf[0]->ms(), "001001000000001");
-	ms1 = bts_ms_store(bts)->get_ms(GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, "001001000000001");
+	ms1 = bts_get_ms(bts, GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, "001001000000001");
 	OSMO_ASSERT(ms1 != NULL);
-	ms2 = bts_ms_store(bts)->get_ms(0xf1000001);
+	ms2 = bts_get_ms_by_tlli(bts, 0xf1000001, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms2 != NULL);
 	OSMO_ASSERT(strcmp(ms_imsi(ms2), "001001000000001") == 0);
 	OSMO_ASSERT(ms1 == ms2);
 
 	/* change the IMSI on TBF 0 */
 	ms_set_imsi(dl_tbf[0]->ms(), "001001000000002");
-	ms1 = bts_ms_store(bts)->get_ms(GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, "001001000000001");
+	ms1 = bts_get_ms(bts, GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, "001001000000001");
 	OSMO_ASSERT(ms1 == NULL);
-	ms1 = bts_ms_store(bts)->get_ms(GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, "001001000000002");
+	ms1 = bts_get_ms(bts, GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, "001001000000002");
 	OSMO_ASSERT(ms1 != NULL);
 	OSMO_ASSERT(strcmp(ms_imsi(ms2), "001001000000002") == 0);
 	OSMO_ASSERT(ms1 == ms2);
 
 	/* use the same IMSI on TBF 1 */
 	{
-		ms_ref(ms2);
+		ms_ref(ms2, __func__);
 		ms_set_imsi(dl_tbf[1]->ms(), "001001000000002");
-		ms1 = bts_ms_store(bts)->get_ms(GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, "001001000000002");
+		ms1 = bts_get_ms(bts, GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, "001001000000002");
 		OSMO_ASSERT(ms1 != NULL);
 		OSMO_ASSERT(ms1 != ms2);
 		OSMO_ASSERT(strcmp(ms_imsi(ms1), "001001000000002") == 0);
 		OSMO_ASSERT(strcmp(ms_imsi(ms2), "") == 0);
-		ms_unref(ms2);
+		ms_unref(ms2, __func__);
 	}
 
-	ms2 = bts_ms_store(bts)->get_ms(0xf1000001);
+	ms2 = bts_get_ms_by_tlli(bts, 0xf1000001, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms2 == NULL);
 
 	tbf_free(dl_tbf[1]);
-	ms1 = bts_ms_store(bts)->get_ms(GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, "001001000000002");
+	ms1 = bts_get_ms(bts, GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, "001001000000002");
 	OSMO_ASSERT(ms1 == NULL);
 
 	TALLOC_FREE(the_pcu);
@@ -567,7 +578,7 @@ static void test_tbf_dl_llc_loss()
 		delay_csec, buf, sizeof(buf));
 	OSMO_ASSERT(rc >= 0);
 
-	ms = bts_ms_store(bts)->get_ms(GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, imsi);
+	ms = bts_get_ms(bts, GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, imsi);
 	OSMO_ASSERT(ms != NULL);
 	OSMO_ASSERT(ms_dl_tbf(ms) != NULL);
 	ms_dl_tbf(ms)->set_ta(0);
@@ -626,7 +637,7 @@ static void test_tbf_dl_llc_loss()
 	fprintf(stderr, "=== end %s ===\n", __func__);
 
 	/* Restore MS release timeout to 0 to make sure it is freed immediately: */
-	ms_set_timeout(ms, 0);
+	OSMO_ASSERT(osmo_tdef_set(the_pcu->T_defs, -2030, 0, OSMO_TDEF_S) == 0);
 	TALLOC_FREE(the_pcu);
 }
 
@@ -639,10 +650,18 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_single_phase(struct gprs_rlcmac_bts
 
 	tfi = bts_tfi_find_free(bts, GPRS_RLCMAC_UL_TBF, &pdch->trx->trx_no, -1);
 
+	/* We set X2002 timeout to 0 here to get immediate trigger through osmo_select_main() below */
+	OSMO_ASSERT(osmo_tdef_set(the_pcu->T_defs, -2002, 0, OSMO_TDEF_MS) == 0);
 	bts_handle_rach(bts, 0x03, *fn, qta);
 
 	ul_tbf = bts_ul_tbf_by_tfi(bts, tfi, pdch->trx->trx_no, pdch->ts_no);
 	OSMO_ASSERT(ul_tbf != NULL);
+	OSMO_ASSERT(ul_tbf->state_is(TBF_ST_ASSIGN));
+
+	/* ImmAss has been sent PCU=PCUIF=>BTS. This means UL TBF is in state
+	 * ASSIGN with X2002 armed. It will move to FLOW once it expires. */
+	osmo_select_main(0);
+	OSMO_ASSERT(ul_tbf->state_is(TBF_ST_FLOW));
 
 	OSMO_ASSERT(ul_tbf->ta() == qta / 4);
 
@@ -656,7 +675,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_single_phase(struct gprs_rlcmac_bts
 
 	pdch->rcv_block(&data_msg[0], sizeof(data_msg), *fn, &meas);
 
-	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms != NULL);
 
 	return ul_tbf;
@@ -704,7 +723,7 @@ static uint32_t get_poll_fn(struct gprs_rlcmac_tbf *tbf, struct gprs_rlcmac_pdch
 
 static void send_control_ack(gprs_rlcmac_tbf *tbf)
 {
-	RlcMacUplink_t ulreq = {0};
+	RlcMacUplink_t ulreq = { };
 
 	ulreq.u.MESSAGE_TYPE = MT_PACKET_CONTROL_ACK;
 	Packet_Control_Acknowledgement_t *ctrl_ack =
@@ -730,7 +749,7 @@ static gprs_rlcmac_ul_tbf *puan_urbb_len_issue(struct gprs_rlcmac_bts *bts,
 	uint32_t sba_fn = *fn + 52;
 	int tfi = 0;
 	gprs_rlcmac_ul_tbf *ul_tbf;
-	RlcMacUplink_t ulreq = {0};
+	RlcMacUplink_t ulreq = { };
 	struct gprs_rlc_ul_header_egprs_3 *egprs3  = NULL;
 
 	/* needed to set last_rts_fn in the PDCH object */
@@ -795,7 +814,7 @@ static gprs_rlcmac_ul_tbf *puan_urbb_len_issue(struct gprs_rlcmac_bts *bts,
 
 	pdch->rcv_block(&data_msg[0], 23, *fn, &meas);
 
-	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms != NULL);
 	OSMO_ASSERT(ms_ta(ms) == qta/4);
 	OSMO_ASSERT(ms_ul_tbf(ms) == ul_tbf);
@@ -873,7 +892,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_spb(struct gprs_rlcmac_bts
 	uint32_t sba_fn = *fn + 52;
 	int tfi = 0, i = 0;
 	gprs_rlcmac_ul_tbf *ul_tbf;
-	RlcMacUplink_t ulreq = {0};
+	RlcMacUplink_t ulreq = { };
 	struct gprs_rlc_ul_header_egprs_3 *egprs3  = NULL;
 
 	/* needed to set last_rts_fn in the PDCH object */
@@ -939,7 +958,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_spb(struct gprs_rlcmac_bts
 
 	pdch->rcv_block(&data_msg[0], 23, *fn, &meas);
 
-	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms != NULL);
 	OSMO_ASSERT(ms_ta(ms) == qta/4);
 	OSMO_ASSERT(ms_ul_tbf(ms) == ul_tbf);
@@ -1304,7 +1323,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf(struct gprs_rlcmac_bts *bts,
 	uint32_t sba_fn = *fn + 52;
 	int tfi = 0;
 	gprs_rlcmac_ul_tbf *ul_tbf;
-	RlcMacUplink_t ulreq = {0};
+	RlcMacUplink_t ulreq = { };
 
 	/* needed to set last_rts_fn in the PDCH object */
 	request_dl_rlc_block(bts, pdch, fn);
@@ -1428,7 +1447,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_puan_URBB_no_length(struct
 	check_tbf(ul_tbf);
 	OSMO_ASSERT(tbf_ul_ack_fi(ul_tbf)->state == TBF_UL_ACK_ST_NONE);
 
-	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms != NULL);
 	OSMO_ASSERT(ms_ta(ms) == qta/4);
 	OSMO_ASSERT(ms_ul_tbf(ms) == ul_tbf);
@@ -1509,7 +1528,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_puan_URBB_with_length(stru
 	check_tbf(ul_tbf);
 	OSMO_ASSERT(tbf_ul_ack_fi(ul_tbf)->state == TBF_UL_ACK_ST_NONE);
 
-	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms != NULL);
 	OSMO_ASSERT(ms_ta(ms) == qta/4);
 	OSMO_ASSERT(ms_ul_tbf(ms) == ul_tbf);
@@ -1591,7 +1610,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase_puan_CRBB(struct gprs_rlcm
 	check_tbf(ul_tbf);
 	OSMO_ASSERT(tbf_ul_ack_fi(ul_tbf)->state == TBF_UL_ACK_ST_NONE);
 
-	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms != NULL);
 	OSMO_ASSERT(ms_ta(ms) == qta/4);
 	OSMO_ASSERT(ms_ul_tbf(ms) == ul_tbf);
@@ -1607,7 +1626,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase(struct gprs_rlcmac_bts *bt
 	uint32_t sba_fn = *fn + 52;
 	int tfi = 0;
 	gprs_rlcmac_ul_tbf *ul_tbf;
-	RlcMacUplink_t ulreq = {0};
+	RlcMacUplink_t ulreq = { };
 
 	/* needed to set last_rts_fn in the PDCH object */
 	request_dl_rlc_block(bts, pdch, fn);
@@ -1668,7 +1687,7 @@ static gprs_rlcmac_ul_tbf *establish_ul_tbf_two_phase(struct gprs_rlcmac_bts *bt
 
 	pdch->rcv_block(&data_msg[0], sizeof(data_msg), *fn, &meas);
 
-	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms != NULL);
 	OSMO_ASSERT(ms_ta(ms) == qta/4);
 	OSMO_ASSERT(ms_ul_tbf(ms) == ul_tbf);
@@ -1681,16 +1700,16 @@ static void send_dl_data(struct gprs_rlcmac_bts *bts, uint32_t tlli, const char
 {
 	GprsMs *ms, *ms2;
 
-	ms = bts_ms_store(bts)->get_ms(tlli, GSM_RESERVED_TMSI, imsi);
+	ms = bts_get_ms(bts, tlli, GSM_RESERVED_TMSI, imsi);
 
 	dl_tbf_handle(bts, tlli, 0, imsi, 0, 0,
 		1000, data, data_size);
 
-	ms = bts_ms_by_imsi(bts, imsi);
+	ms = bts_get_ms_by_imsi(bts, imsi);
 	OSMO_ASSERT(ms != NULL);
 
 	if (imsi[0] != '\0') {
-		ms2 = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
+		ms2 = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 		OSMO_ASSERT(ms == ms2);
 	}
 }
@@ -1702,7 +1721,7 @@ static void transmit_dl_data(struct gprs_rlcmac_bts *bts, uint32_t tlli, uint32_
 	GprsMs *ms;
 	unsigned ts_no;
 
-	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms);
 	dl_tbf = ms_dl_tbf(ms);
 	OSMO_ASSERT(dl_tbf);
@@ -2011,14 +2030,14 @@ static void test_tbf_ra_update_rach()
 	 * the PCU can see, that both MS objects belong to same MS */
 	send_dl_data(bts, tlli2, imsi, (const uint8_t *)"DATA", 4);
 
-	ms = bts_ms_by_imsi(bts, imsi);
+	ms = bts_get_ms_by_imsi(bts, imsi);
 	OSMO_ASSERT(ms == ms2);
 
 	print_ms(ms2, false);
 
-	ms = bts_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms == NULL);
-	ms = bts_ms_by_tlli(bts, tlli2, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli2, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms == ms2);
 
 	TALLOC_FREE(the_pcu);
@@ -2061,7 +2080,7 @@ static void test_tbf_dl_flow_and_rach_two_phase()
 
 	/* Get rid of old UL TBF */
 	tbf_free(ul_tbf);
-	ms = bts_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms1 == ms);
 
 	/* Now establish a new UL TBF, this will consume one LLC packet */
@@ -2074,7 +2093,7 @@ static void test_tbf_dl_flow_and_rach_two_phase()
 	/* This should be the same MS object */
 	OSMO_ASSERT(ms2 == ms1);
 
-	ms = bts_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms2 == ms);
 
 	/* A DL TBF should still exist */
@@ -2124,7 +2143,7 @@ static void test_tbf_dl_flow_and_rach_single_phase()
 
 	/* Get rid of old UL TBF */
 	tbf_free(ul_tbf);
-	ms = bts_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms1 == ms);
 
 	/* Now establish a new UL TBF */
@@ -2136,7 +2155,7 @@ static void test_tbf_dl_flow_and_rach_single_phase()
 	/* There should be a different MS object */
 	OSMO_ASSERT(ms2 != ms1);
 
-	ms = bts_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms2 == ms);
 	OSMO_ASSERT(ms1 != ms);
 
@@ -2165,7 +2184,7 @@ static void test_tbf_dl_reuse()
 	gprs_rlcmac_dl_tbf *dl_tbf1, *dl_tbf2;
 	GprsMs *ms1, *ms2;
 	unsigned i;
-	RlcMacUplink_t ulreq = {0};
+	RlcMacUplink_t ulreq = { };
 
 	fprintf(stderr, "=== start %s ===\n", __func__);
 
@@ -2233,7 +2252,7 @@ static void test_tbf_dl_reuse()
 
 	request_dl_rlc_block(dl_tbf1, &fn);
 
-	ms2 = bts_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
+	ms2 = bts_get_ms_by_tlli(bts, tlli1, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms2 == ms1);
 	OSMO_ASSERT(ms_dl_tbf(ms2));
 	OSMO_ASSERT(ms_dl_tbf(ms2)->state_is(TBF_ST_ASSIGN));
@@ -2345,16 +2364,23 @@ static void test_tbf_ws()
 	gprs_bssgp_init(bts, 4234, 4234, 1, 1, false, 0, 0, 0);
 
 	/* Does no support EGPRS */
-	ms = bts_alloc_ms(bts, ms_class, 0);
-	dl_tbf = dl_tbf_alloc(bts, ms, 0, false);
+	ms = ms_alloc(bts, NULL);
+	ms_set_ms_class(ms, ms_class);
+	OSMO_ASSERT(ms_new_dl_tbf_assigned_on_pch(ms) == 0);
+	dl_tbf = ms_dl_tbf(ms);
+	OSMO_ASSERT(dl_tbf);
 
 	ws_check(dl_tbf, __func__, 4, 64, true);
 
 	/* EGPRS-only */
 
 	/* Does support EGPRS */
-	ms = bts_alloc_ms(bts, ms_class, ms_class);
-	dl_tbf = dl_tbf_alloc(bts, ms, 0, false);
+	ms = ms_alloc(bts, NULL);
+	ms_set_ms_class(ms, ms_class);
+	ms_set_egprs_ms_class(ms, ms_class);
+	OSMO_ASSERT(ms_new_dl_tbf_assigned_on_pch(ms) == 0);
+	dl_tbf = ms_dl_tbf(ms);
+	OSMO_ASSERT(dl_tbf);
 
 	ws_check(dl_tbf, __func__, 4, 128 + 4 * 64, true);
 	fprintf(stderr, "=== end %s ===\n", __func__);
@@ -2393,8 +2419,12 @@ static void test_tbf_update_ws(void)
 	/* EGPRS-only */
 
 	/* Does support EGPRS */
-	ms = bts_alloc_ms(bts, ms_class, ms_class);
-	dl_tbf = dl_tbf_alloc(bts, ms, 0, true);
+	ms = ms_alloc(bts, NULL);
+	ms_set_ms_class(ms, ms_class);
+	ms_set_egprs_ms_class(ms, ms_class);
+	OSMO_ASSERT(ms_new_dl_tbf_assigned_on_pch(ms) == 0);
+	dl_tbf = ms_dl_tbf(ms);
+	OSMO_ASSERT(dl_tbf);
 
 	ws_check(dl_tbf, __func__, 1, 128 + 1 * 64, false);
 
@@ -2457,7 +2487,7 @@ static void test_ms_merge_dl_tbf_different_trx(void)
 			   delay_csec, llc_buf, sizeof(llc_buf));
 	OSMO_ASSERT(rc >= 0);
 
-	first_ms = bts_ms_store(bts)->get_ms(GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, imsi);
+	first_ms = bts_get_ms(bts, GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, imsi);
 	OSMO_ASSERT(first_ms);
 	dl_tbf = ms_dl_tbf(first_ms);
 	OSMO_ASSERT(dl_tbf);
@@ -2470,9 +2500,10 @@ static void test_ms_merge_dl_tbf_different_trx(void)
 	trx0->pdch[2].enable();
 	trx0->pdch[3].enable();
 
-	second_ms = bts_alloc_ms(bts, 0, 0);
+	second_ms = ms_alloc(bts, NULL);
 	ms_set_tlli(second_ms, new_tlli);
-	ul_tbf = ul_tbf_alloc(bts, second_ms, 0, true);
+	OSMO_ASSERT(ms_new_ul_tbf_assigned_pacch(second_ms, 0));
+	ul_tbf = ms_ul_tbf(second_ms);
 	OSMO_ASSERT(ul_tbf != NULL);
 	ms_update_announced_tlli(second_ms, new_tlli);
 
@@ -2494,7 +2525,7 @@ static void test_ms_merge_dl_tbf_different_trx(void)
 	osmo_fsm_inst_dispatch(ms_dl_tbf(second_ms)->state_fi, TBF_EV_ASSIGN_PCUIF_CNF, NULL);
 	osmo_select_main(0);
 
-	/* get the PACCH PktDlAss for the DL-TBF, allocated one the UL-TBF from the new MS obj: */
+	/* get the PACCH PktDlAss for the DL-TBF, allocated on the UL-TBF from the new MS obj: */
 	request_dl_rlc_block(dl_tbf->bts, dl_tbf->control_ts, &fn);
 
 	fprintf(stderr, "=== end %s ===\n", __func__);
@@ -2543,7 +2574,7 @@ static gprs_rlcmac_ul_tbf *tbf_li_decoding(struct gprs_rlcmac_bts *bts,
 	uint32_t sba_fn = *fn + 52;
 	int tfi = 0;
 	gprs_rlcmac_ul_tbf *ul_tbf;
-	RlcMacUplink_t ulreq = {0};
+	RlcMacUplink_t ulreq = { };
 	struct gprs_rlc_ul_header_egprs_3 *egprs3  = NULL;
 	Packet_Resource_Request_t *presreq = NULL;
 	MS_Radio_Access_capability_t *pmsradiocap = NULL;
@@ -2600,7 +2631,7 @@ static gprs_rlcmac_ul_tbf *tbf_li_decoding(struct gprs_rlcmac_bts *bts,
 
 	uint8_t data_msg[49] = {0};
 
-	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
+	ms = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	OSMO_ASSERT(ms != NULL);
 	OSMO_ASSERT(ms_ta(ms) == qta/4);
 	OSMO_ASSERT(ms_ul_tbf(ms) == ul_tbf);
@@ -2882,9 +2913,16 @@ static void establish_and_use_egprs_dl_tbf(struct gprs_rlcmac_bts *bts, int mcs)
 		/* Request to send one RLC/MAC block */
 		request_dl_rlc_block(dl_tbf, &fn);
 	}
+	/* ACK all blocks */
+	memset(rbb, 0xff, sizeof(rbb));
+	_rcv_ack(false, dl_tbf, rbb);
+	/* X2031 may keep the TBF open for a while: */
+	while (dl_tbf->state_is(TBF_ST_FLOW)) {
+		request_dl_rlc_block(dl_tbf, &fn);
+	}
 	send_empty_block(dl_tbf, dl_tbf->control_ts, fn);
 
-	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FLOW));
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FINISHED));
 
 	_rcv_ack(true, dl_tbf, rbb); /* Receive a final ACK */
 
@@ -2933,6 +2971,19 @@ static gprs_rlcmac_dl_tbf *tbf_init(struct gprs_rlcmac_bts *bts,
 static void tbf_cleanup(gprs_rlcmac_dl_tbf *dl_tbf)
 {
 	uint8_t rbb[64/8];
+	uint32_t fn = 0;
+
+	if (dl_tbf->state_is(TBF_ST_FLOW)) {
+		/* X2031 may keep the TBF open for a while: */
+		while (dl_tbf->state_is(TBF_ST_FLOW)) {
+			/* ACK all blocks */
+			memset(rbb, 0xff, sizeof(rbb));
+			_rcv_ack(false, dl_tbf, rbb);
+			request_dl_rlc_block(dl_tbf, &fn);
+		}
+	}
+
+	OSMO_ASSERT(dl_tbf->state_is(TBF_ST_FINISHED));
 
 	_rcv_ack(true, dl_tbf, rbb); /* Receive a final ACK */
 
@@ -3224,7 +3275,6 @@ static void establish_and_use_egprs_dl_tbf_for_retx(struct gprs_rlcmac_bts *bts,
 		MAKE_ACKED(msg, dl_tbf, fn, mcs, true);
 	}
 	/* Clean up pending items in UL controller: */
-	send_empty_block(dl_tbf, dl_tbf->control_ts, fn+50);
 	tbf_cleanup(dl_tbf);
 }
 
@@ -3329,7 +3379,7 @@ static void test_packet_access_rej_prr_no_other_tbfs()
 
 	int rc = 0;
 
-	ms = bts_alloc_ms(bts, 0, 0);
+	ms = ms_alloc(bts, NULL);
 	ms_set_tlli(ms, tlli);
 	ul_tbf = ms_new_ul_tbf_rejected_pacch(ms, pdch);
 
@@ -3356,7 +3406,7 @@ static void test_packet_access_rej_prr()
 	uint16_t qta = 31;
 	int ts_no = 7;
 	uint8_t trx_no = 0;
-	RlcMacUplink_t ulreq = {0};
+	RlcMacUplink_t ulreq = { };
 	Packet_Resource_Request_t *presreq = NULL;
 	uint8_t ms_class = 11;
 	uint8_t egprs_ms_class = 11;
@@ -3450,7 +3500,7 @@ void test_packet_access_rej_epdan()
 
 int main(int argc, char **argv)
 {
-	struct vty_app_info pcu_vty_info = {0};
+	struct vty_app_info pcu_vty_info = { };
 
 	tall_pcu_ctx = talloc_named_const(NULL, 1, "moiji-mobile TbfTest context");
 	if (!tall_pcu_ctx)
@@ -3464,7 +3514,7 @@ int main(int argc, char **argv)
 	log_set_print_category_hex(osmo_stderr_target, 0);
 	log_parse_category_mask(osmo_stderr_target, "DRLCMAC,1:DRLCMACDATA,3:DRLCMACDL,3:DRLCMACUL,3:"
 				"DRLCMACSCHED,1:DRLCMACMEAS,3:DNS,3:DLBSSGP,3:DPCU,5:"
-				"DL1IF,6:DTBF,1:DTBFUL,1:DTBFDL,1:DLGLOBAL,2:");
+				"DL1IF,6:DMS,1:DTBF,1:DTBFUL,1:DTBFDL,1:DLGLOBAL,2:");
 	osmo_fsm_log_addr(false);
 	vty_init(&pcu_vty_info);
 	pcu_vty_init();
@@ -3517,6 +3567,7 @@ int main(int argc, char **argv)
 extern "C" {
 void l1if_pdch_req() { abort(); }
 void l1if_connect_pdch() { abort(); }
-void l1if_close_pdch() { abort(); }
-void l1if_open_pdch() { abort(); }
+void l1if_disconnect_pdch() { abort(); }
+void l1if_close_trx() { abort(); }
+void l1if_open_trx() { abort(); }
 }
